(* FlatCopy
   ========
   Copy files from a directory tree into one single directory

    Dr. J. Rathlev, D-24222 Schwentinental (kontakt(a)rathlev-home.de)

   The contents of this file may be used under the terms of the
   Mozilla Public License ("MPL") or
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   J. Rathlev, Jan. 2008
   last modified: February 2023
   *)

unit FlatCopyMain;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes,
  Vcl.Graphics, Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.Buttons, Vcl.StdCtrls,
  HListBox, Vcl.ComCtrls, Vcl.ExtCtrls, NumberEd, TimeSlotDlg;

const
  ProgName = 'Flat file copy';
  Vers = '2.1';
  CopRgt = ' 2008-2023 Dr. J. Rathlev, 24222 Schwentinental';
  EmailAdr = 'kontakt(a)rathlev-home.de';

type
  TFileData = class (TObject)
    Path : string;
    FTs  : TDateTime;
    Dup  : integer;
    constructor Create (const APath : string; ATs : TDateTime);
    procedure Update (const APath : string; ATs : TDateTime);
    end;

  TMainForm = class(TForm)
    Label1: TLabel;
    btSource: TSpeedButton;
    Label2: TLabel;
    Label3: TLabel;
    btTarget: TSpeedButton;
    btFilter: TSpeedButton;
    cbSubDir: TCheckBox;
    rgOverwrite: TRadioGroup;
    btnQuit: TBitBtn;
    StatusBar: TStatusBar;
    btCopy: TBitBtn;
    lbFiles: TListBox;
    Label4: TLabel;
    cbSync: TCheckBox;
    cbSourceDelete: TCheckBox;
    pcOptions: TPageControl;
    tsOptions: TTabSheet;
    tsTime: TTabSheet;
    Label5: TLabel;
    reTimeDiff: TRangeEdit;
    udTimeDiff: TNumUpDown;
    cbDstIgnore: TCheckBox;
    tsDuplicates: TTabSheet;
    rbRetainRecent: TRadioButton;
    rbRetainAll: TRadioButton;
    rbRetainOldest: TRadioButton;
    edSource: TComboBox;
    edFilter: TComboBox;
    edDest: TComboBox;
    btInfo: TBitBtn;
    Label6: TLabel;
    edTimeSlot: TEdit;
    btTimeSlot: TSpeedButton;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure btnQuitClick(Sender: TObject);
    procedure btnInfoClick(Sender: TObject);
    procedure btSourceClick(Sender: TObject);
    procedure btTargetClick(Sender: TObject);
    procedure btFilterClick(Sender: TObject);
    procedure btCopyClick(Sender: TObject);
    procedure rgOverwriteClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure btTimeSlotClick(Sender: TObject);
    procedure edFilterCloseUp(Sender: TObject);
    procedure edSourceCloseUp(Sender: TObject);
    procedure edDestCloseUp(Sender: TObject);
  private
    { Private-Deklarationen }
    ProgVersName,
    ProgVersDate,
    ProgVers,
    AppPath,UserPath,
    IniName             : string;
    Running             : boolean;
    TmPeriod            : TTimePeriod;
  public
    { Public-Deklarationen }
  end;

var
  MainForm: TMainForm;

implementation

{$R *.dfm}

uses ShellDirDlg, Winapi.ShlObj, System.IniFiles, InitProg, FileUtils, PathUtils,
  WinUtils, MsgDialogs, GnuGetText, WinShell, LangUtils, StringUtils, ExtSysUtils,
  SelectFromListDlg;

constructor TFileData.Create (const APath : string; ATs : TDateTime);
begin
  inherited Create;
  Path:=APath; FTs:=ATs;  Dup:=0;
  end;

procedure TFileData.Update (const APath : string; ATs : TDateTime);
begin
  Path:=APath; FTs:=ATs;
  end;

const
  IniExt = 'ini';

  (* INI-Sektionen *)
  CfGSekt = 'Config';
  SourceSekt = 'Source';
  DestSekt = 'Dest';
  FilterSekt = 'Filter';

  (* INI-Variablen *)
  iniSubDir = 'Subdirectories';
  iniOvrWrt = 'Overwrite';
  iniTmDiff = 'TimeMargin';
  iniDst    = 'IgnoreDst';
  iniDelSrc = 'DeleteSource';
  iniDupl   = 'Duplicates';
  iniTmRef     = 'TimeRef';
  iniTmMode    = 'TimeMode';
  iniDayTime   = 'DayTime';
  iniDays      = 'Days';
  iniFrom      = 'DateFrom';
  iniTo        = 'DateTo';

procedure TMainForm.FormCreate(Sender: TObject);
var
  n : integer;
  t : string;
begin
  TranslateComponent(self);
  InitPaths(AppPath,UserPath);
  InitVersion(ProgName,Vers,CopRgt,3,3,ProgVersName,ProgVers,ProgVersDate);
  Caption:=_('Copy files into one folder');
  with TmPeriod do begin
    Mode:=tmAll; TmRef:=trModify;
    Dat1:=Date-1461; Dat2:=Date;   // 4 Jahre
    DayTime:=0.25; Days:=7;
    end;
  IniName:=Erweiter(AppPath,PrgName,IniExt);
  with TIniFile.Create(IniName) do begin
    cbSubDir.Checked:=ReadBool(CfgSekt,iniSubDir,true);
    rgOverwrite.ItemIndex:=ReadInteger(CfgSekt,iniOvrWrt,0);
    reTimeDiff.Value:=ReadInteger(CfgSekt,iniTmDiff,2);
    cbDstIgnore.Checked:=ReadBool(CfgSekt,iniDst,false);
    cbSourceDelete.Checked:=ReadBool(CfgSekt,iniDelSrc,false);
    n:=ReadInteger(CfgSekt,iniDupl,0);
    case n of
    0 : rbRetainRecent.Checked:=true;
    1 : rbRetainOldest.Checked:=true;
    else rbRetainAll.Checked:=true;
      end;
    with TmPeriod do begin
      Days:=ReadInteger(CfgSekt,iniDays,7);
      Mode:=TTimeMode(ReadInteger(CfgSekt,iniTmMode,0));
      TmRef:=TTimeRef(ReadInteger(CfgSekt,iniTmRef,0));
      if Mode>tmUpdate then Mode:=tmAll;
      try
        t:=ReadString(CfgSekt,iniDayTime,'06'+FormatSettings.TimeSeparator+'00');
        DayTime:=StrToTime(t);
        t:=ReadString(CfgSekt,iniFrom,DateToStr(Date-10));
        Dat1:=StrToDate(t);
        t:=ReadString(CfgSekt,iniTo,DateToStr(Date));
        Dat2:=StrToDate(t);
      except
        Mode:=tmAll;
        end;
      end;
    Free;
    end;
  with edSource do begin
    LoadHistory(IniName,SourceSekt,'',Items);
    ItemIndex:=0;
    end;
  with edFilter do begin
    LoadHistory(IniName,FilterSekt,'',Items);
    ItemIndex:=0;
    end;
  with edDest do begin
    LoadHistory(IniName,DestSekt,'',Items);
    ItemIndex:=0;
    end;
  pcOptions.ActivePageIndex:=0;
  tsTime.Tabvisible:=rgOverwrite.ItemIndex=1;
  Running:=false;
  end;

procedure TMainForm.FormDestroy(Sender: TObject);
var
  n : Integer;
begin
  with TIniFile.Create(IniName)do begin
    WriteBool(CfgSekt,iniSubDir,cbSubDir.Checked);
    WriteInteger(CfgSekt,iniOvrWrt,rgOverwrite.ItemIndex);
    WriteInteger(CfgSekt,iniTmDiff,reTimeDiff.Value);
    WriteBool(CfgSekt,iniDst,cbDstIgnore.Checked);
    WriteBool(CfgSekt,iniDelSrc,cbSourceDelete.Checked);
    if rbRetainRecent.Checked then n:=0
    else if rbRetainOldest.Checked then n:=1
    else n:=2;
    WriteInteger(CfgSekt,iniDupl,n);
    with TmPeriod do begin
      WriteInteger(CfgSekt,iniTmMode,integer(Mode));
      WriteInteger(CfgSekt,iniTmRef,integer(TmRef));
      WriteInteger(CfgSekt,iniDays,Days);
      WriteString(CfgSekt,iniDayTime,TimeToStr(DayTime));
      WriteString(CfgSekt,iniFrom,DateToStr(Dat1));
      WriteString(CfgSekt,iniTo,DateToStr(Dat2));
      end;
    Free;
    end;
  with edSource do SaveHistory(IniName,SourceSekt,'',true,Items,DropDownCount);
  with edFilter do SaveHistory(IniName,FilterSekt,'',true,Items,DropDownCount);
  with edDest do SaveHistory(IniName,DestSekt,'',true,Items,DropDownCount);
  end;

procedure TMainForm.FormShow(Sender: TObject);
begin
  edTimeSlot.Text:=GetTimeSlotAsString(TmPeriod);
  end;

procedure TMainForm.rgOverwriteClick(Sender: TObject);
begin
  tsTime.Tabvisible:=rgOverwrite.ItemIndex=1;
  end;

procedure TMainForm.btnInfoClick(Sender: TObject);
begin
  InfoDialog (TopRightPos(btFilter,5,0),Caption+ProgVers+' - '+ProgVersDate+sLineBreak+CopRgt
           +sLineBreak+'E-Mail: '+EmailAdr);
  end;

procedure TMainForm.btnQuitClick(Sender: TObject);
begin
  Close;
  end;

procedure TMainForm.btSourceClick(Sender: TObject);
var
  s : string;
begin
  s:=edSource.Text;
  if length(s)=0 then s:=UserPath;
  if ShellDirDialog.Execute (_('Select directory with files to be copied'),true,true,false,'',s) then begin
    with edSource do begin
      Text:=s; AddToHistory(Items,s,DropDownCount);
      end;
    with edDest do if length(Text)=0 then begin
      Text:=s+'\flat'; AddToHistory(Items,Text,DropDownCount);
      end;
    with edFilter do begin
      Text:='*.*'; AddToHistory(Items,Text,DropDownCount);
      end;
    end;
  end;

procedure TMainForm.edSourceCloseUp(Sender: TObject);
begin
  with edSource do begin
    AddToHistory(Items,Items[ItemIndex],DropDownCount);
    ItemIndex:=0;
    end;
  end;

procedure TMainForm.btFilterClick(Sender: TObject);
var
  s  : string;
  ml : TStringList;
begin
  s:='';
  ml:=TStringList.Create;
  with ml do begin
    Delimiter:=';';
    DelimitedText:=edFilter.Text;
    end;
  if SelectFromListDialog.Select(TopRightPos(btFilter,5,0),
              Caption,_('File filter:'),'',
              [soEdit,soOrder],0,tcLower,'*.*',ml,s) then begin
    with edFilter do begin
      Text:=ml.DelimitedText; AddToHistory(Items,Text,DropDownCount);
      end;
    ml.Free;
    end;
  end;

procedure TMainForm.edFilterCloseUp(Sender: TObject);
begin
  with edFilter do begin
    AddToHistory(Items,Items[ItemIndex],DropDownCount);
    ItemIndex:=0;
    end;
  end;

procedure TMainForm.btTimeSlotClick(Sender: TObject);
begin
  if TimeSlotDialog.Execute(TopRightPos(btTimeSlot,5,0),TmPeriod) then
    edTimeSlot.Text:=GetTimeSlotAsString(TmPeriod);
  end;

procedure TMainForm.btTargetClick(Sender: TObject);
var
  s : string;
begin
  s:=edDest.Text;
  if length(s)=0 then s:=edSource.Text;
  if length(s)=0 then s:=UserPath;
  if ShellDirDialog.Execute(_('Select target directory for files to be copied'),true,true,false,'',s) then begin
    with edDest do begin
      Text:=s; AddToHistory(Items,Text,DropDownCount);
      end;
    end;
  end;

procedure TMainForm.edDestCloseUp(Sender: TObject);
begin
  with edDest do begin
    AddToHistory(Items,Items[ItemIndex],DropDownCount);
    ItemIndex:=0;
    end;
  end;

procedure TMainForm.btCopyClick(Sender: TObject);
var
  FListe,
  SListe     : TStringList;
  s,t,sd,ss  : string;
  dft,ec,
  i,n,m,k    : integer;
  ok         : boolean;

  function CheckTime (const tp : TTimePeriod; const DirInfo : TSearchRec) : boolean;
  var
    fi : TFileInfo;

    function MatchesTime(dt : TDateTime) : boolean;
    begin
      with tp do Result:=(Mode=tmAll) or
          ((Mode=tmNewer) and (Dat1<=dt)) or
          ((Mode=tmOlder) and (Dat2>dt)) or
          ((Mode=tmSpan) and (Dat1<=dt) and (dt<Dat2+1)) or
          ((Mode=tmToday) and (Date+DayTime<=dt)) or
          ((Mode=tmPrevDays) and (Date-Days>int(dt))) or
          ((Mode=tmLastDays) and (Date-Days<=int(dt)));
      end;

  begin
    fi:=SearchRecToFileInfo(DirInfo);
    if tp.TmRef<>trCreate then begin
      Result:=MatchesTime(fi.DateTime);
      if not Result and (tp.TmRef=trBoth) then Result:=MatchesTime(fi.CreationTime);
      end
    else Result:=MatchesTime(fi.CreationTime);
    end;

  procedure ScanDir (Dir,Mask : string);
  var
    DirInfo    : TSearchRec;
    Findresult,n : integer;
  begin
    StatusBar.SimpleText:=_('Scanning: ')+StripPath(Dir,50);
    Application.ProcessMessages;
    if cbSubDir.Checked then begin
      FindResult:=FindFirst (Erweiter(Dir,'*','*'),faAnyFile,DirInfo);
      while FindResult=0 do with DirInfo do begin
        if NotSpecialDir(Name) then begin
          if((Attr and faDirectory)<>0) then ScanDir (Erweiter(Dir,Name,''),Mask);
          end;
        FindResult:=FindNext (DirInfo);
        end;
      FindClose(DirInfo);
      end;
    FindResult:=FindFirst (Erweiter(Dir,Mask,''),faArchive+faReadOnly+faHidden+faSysfile,DirInfo);
    while FindResult=0 do with FListe do begin
      if NotSpecialDir(Name) and CheckTime(TmPeriod,DirInfo) then begin
        n:=FListe.IndexOf(DirInfo.Name);
        if n<0 then AddObject(DirInfo.Name,TFileData.Create(AddPath(Dir,DirInfo.Name),DirInfo.Timestamp))
        else with (Objects[n] as TFileData) do begin  // already  in list
          if rbRetainRecent.Checked then begin
            if DirInfo.Timestamp>=FTs then Update(AddPath(Dir,DirInfo.Name),DirInfo.Timestamp);
            end
          else if rbRetainOldest.Checked then begin
            if DirInfo.Timestamp<FTs then Update(AddPath(Dir,DirInfo.Name),DirInfo.Timestamp);
            end
          else begin
            inc(Dup);
            AddObject(InsertNameSuffix(DirInfo.Name,'-'+IntToStr(Dup)),
              TFileData.Create(AddPath(Dir,DirInfo.Name),DirInfo.Timestamp));
            end;
          end;
        end;
      FindResult:=FindNext (DirInfo);
      end;
    FindClose(DirInfo);
    end;

  function SyncFiles (FileList : TStringList; const Dir,Mask : string) : integer;
  var
    DirInfo    : TSearchRec;
    Findresult : integer;
    s          : string;
  begin
    Result:=0;
    FindResult:=FindFirst (Erweiter(Dir,Mask,''),faArchive+faReadOnly+faHidden+faSysfile,DirInfo);
    while FindResult=0 do with DirInfo do begin
      if NotSpecialDir(Name) and (FileList.IndexOf(Name)<0) then begin
        s:=Erweiter(Dir,Name,'');
        EraseFile(s);
        lbFiles.Items.Add('  '+s);
        inc(Result);
        end;
      FindResult:=FindNext (DirInfo);
      end;
    FindClose(DirInfo);
    end;

begin
  if Running then Running:=false
  else if not cbSourceDelete.Checked or
      ConfirmDialog(BottomRightPos(btCopy),
          _('You selected the option "Delete source files after successful copy"!'+sLineBreak+
          'Do you really want to start copying files?')) then begin
    Running:=true;
    btCopy.Caption:=_('Stop');
    t:=edFilter.Text;
    sd:=IncludeTrailingPathDelimiter(edDest.Text);
    n:=0; m:=0;
    if length(t)>0 then begin
      FListe:=TStringList.Create;
      ForceDirectories(sd);
      lbFiles.Clear;
      repeat
        s:=ReadNxtStr (t,';');
        ScanDir (IncludeTrailingPathDelimiter(edSource.Text),s);
        until (length(t)=0) or not Running;
      k:=0;
      if Running and cbSync.Checked then begin  //  synchronize
        StatusBar.SimpleText:=_('Synchronizing');
        Application.ProcessMessages;
        lbFiles.Items.Add(_('Delete files:'));
        SListe:=TStringList.Create;
        with SListe do begin
          Sorted:=true;
          Duplicates:=dupIgnore;
          for i:=0 to FListe.Count-1 do Add(ExtractFileName((FListe.Objects[i] as TFileData).Path));
          t:=edFilter.Text;
          repeat
            s:=ReadNxtStr (t,';');
            k:=k+SyncFiles(SListe,IncludeTrailingPathDelimiter(edDest.Text),s);
            until (length(t)=0) or not Running;
          Free;
          end;
        if k=0 then lbFiles.Items.Add('  '+_('No files'));
        end;
      if Running then begin
        lbFiles.Items.Add(_('Copy files:'));
        with FListe do if Count>0 then for i:=0 to Count-1 do begin
          StatusBar.SimpleText:=_('Copying: ')+Strings[i];
          Application.ProcessMessages;
          s:=sd+Strings[i];
          ss:=(Objects[i] as TFileData).Path;
          case rgOverwrite.ItemIndex of
          1 : begin // Zeitvergleich
              ok:=not FileExists(s);
              if not ok then begin
                dft:=FileTimeToUnixTime(GetFileLastWriteTime(ss))-
                     FileTimeToUnixTime(GetFileLastWriteTime(s));
                ok:=dft>reTimeDiff.Value;
                if cbDstIgnore.Checked then begin
                  if ok then ok:=abs(dft-3600)>reTimeDiff.Value;
                  end;
                end;
              end;
          2 : ok:=true;
          else ok:=not FileExists(s);
            end;
          if ok then begin
            ec:=0;
            try
              CopyFileTS(ss,s);
            except
              on E:ECopyError do begin
                ec:=E.ErrorCode;
                ErrorDialog(_('Error copying file: ')+Strings[i]+sLineBreak+
                  E.Message+sLineBreak+SystemErrorMessage(ec));
                  ok:=false;
                  end
              else ok:=false;
              end;
            if ok then begin
              lbFiles.Items.Add('  '+Strings[i]);
              inc(n);
              if cbSourceDelete.Checked then begin
                if (Win32Platform=VER_PLATFORM_WIN32_NT) and (Win32MajorVersion>=6) then
                  ec:=IShellDeleteFiles(Application.Handle,ss,true,false)
                else ec:=ShellDeleteFiles(Application.Handle,ss,'',true,false);
                if NoError(ec) then lbFiles.Items.Add('  '+_('Source file deleted')+':'+ss)
                else lbFiles.Items.Add('* '+Format(_('Error deleting source file (0x%.8x)'),[ec])+':'+ss);
                end;
              end
            else begin
              lbFiles.Items.Add('* '+Strings[i]+Format(_(' - Error (0x%.8x)'),[ec]));
              lbFiles.Items.Add('*   '+_('Source: ')+ss);
              end;
            end
          else begin
            lbFiles.Items.Add('  * '+_('Skipped: ')+Strings[i]);
            inc(m);
            end;
          if not Running then Break;
          end;
        if n+m=0 then lbFiles.Items.Add('  '+_('No files'));
        end;
      end;
    FreeListObjects(FListe);
    FListe.Free;
    Running:=false;
    if cbSync.Checked then
      StatusBar.SimpleText:=Format(_('Status: %u copied, %u deleted, %u skipped'),[n,k,m])
    else StatusBar.SimpleText:=Format(_('Status: %u copied, %u skipped'),[n,m]);
    btCopy.Caption:=_('Start');
    end;
  end;

end.
