(* Personal Backup
   ===============
   - Einstellungen fr Zeitfenster

    J. Rathlev, D.24222 Schwentinental (pb(a)rathlev-home.de))

   The contents of this file may be used under the terms of the
   Mozilla Public License ("MPL") or
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   J. Rathlev, Feb. 2008
   last modified: Nov. 2020
   *)

unit TimeSlotDlg;

interface

uses Winapi.Windows, System.SysUtils, System.Classes, Vcl.Graphics, Vcl.Forms,
  Vcl.Controls, Vcl.StdCtrls, Vcl.Buttons, Vcl.ExtCtrls, Vcl.ComCtrls, NumberEd;

resourcestring
  rsAll = 'all';
  rsAfter = 'from %s';
  rsBefore = 'to %s';
  rsSpan = 'from %s to %s';
  rsTodayFrom = 'today from %s';
  rsNewerThan = 'not older than %s';
  rsOlderThan = 'older than %s';
  rsDay = 'day';
  rsDays = 'days';

type
  TTimeMode = (tmAll,tmNewer,tmOlder,tmSpan,tmToday,tmLastDays,tmPrevDays,tmUpdate,
               tmOff,tmError,tmPrompt);
  TTimeRef = (trModify,trCreate,trBoth);

 (* Mode = 0 : alle
           1 : neuer als Dat1
           2 : lter als Dat2
           3 : Zeitraum zwischen Dat1 und Dat2
           4 : heute ab DayTime
           5 : die letzten n Tage
           6 : lter als n Tage
           7 : neuer als letztes Backup
           8 : aus
           9 : Fehler
           10 : Dialog *)
  TTimePeriod = record
    Mode      : TTimeMode;
    TmRef     : TTimeRef;  // Referenz fr Zeitvergleich
    Days      : integer;
    DayTime,
    Dat1,Dat2 : TDateTime;
    end;

  TTimeSlotDialog = class(TForm)
    gbTimeSlot: TGroupBox;
    lbDays: TLabel;
    lbFrom: TLabel;
    lbTo: TLabel;
    ToDate: TDateTimePicker;
    FromDate: TDateTimePicker;
    rbTmAll: TRadioButton;
    rbTmNewer: TRadioButton;
    rbTmOlder: TRadioButton;
    rbTmSpan: TRadioButton;
    rbTmToday: TRadioButton;
    rbTmLastDays: TRadioButton;
    OKBtn: TBitBtn;
    CancelBtn: TBitBtn;
    rbTmPrevDays: TRadioButton;
    reDays: TRangeEdit;
    udDays: TNumUpDown;
    rgTimeSelect: TRadioGroup;
    procedure FormCreate(Sender: TObject);
    procedure rbTmClick(Sender: TObject);
    procedure FromDateChange(Sender: TObject);
    procedure ToDateChange(Sender: TObject);
    procedure FormShow(Sender: TObject);
  private
    { Private declarations }
    DpShortdateFmt, TpShortTimeFmt : string;
    procedure ShowTimeFields;
  public
    { Public declarations }
    procedure AfterConstruction; override;
    function Execute (APos : TPoint; var TimeSlot : TTimePeriod) : boolean;
  end;

function GetTimeSlotAsString (const TimeSlot : TTimePeriod) : string;
function EditTimeSlot (APos : TPoint; var TimeSlot : TTimePeriod) : boolean;

var
  TimeSlotDialog: TTimeSlotDialog;

implementation

{$R *.dfm}

uses System.DateUtils, WinUtils, GnuGetText, ExtSysUtils, StringUtils;

{------------------------------------------------------------------- }
procedure TTimeSlotDialog.FormCreate(Sender: TObject);
begin
  TranslateComponent (self);
  // LOCALE_ITIME = 0: AM/PM 12-hour format,
  //                1: 24-hour format
  if StrToIntDef(GetLocaleStr(GetUserDefaultLCID,LOCALE_ITIME,'1'),1)=0 then TpShortTimeFmt:='hh:mm tt'
  else TpShortTimeFmt:='HH:mm';
  // LOCALE_IDATE = 0: Month - Day - Year
	//                1: Day - Month - Year
	//                2: Year - Month - Day
  case StrToIntDef(GetLocaleStr(GetUserDefaultLCID,LOCALE_IDATE,'2'),2) of
  0 : DpShortdateFmt:='MM/dd/yyyy';
  1 : DpShortdateFmt:='dd.MM.yyyy';
  else DpShortdateFmt:='yyyy-MM-dd';
    end;
  FromDate.Format:=DpShortDateFmt;
  ToDate.Format:=DpShortDateFmt;
  end;

procedure TTimeSlotDialog.AfterConstruction;
begin
  inherited;
  if Application.Tag=0 then
    ScaleButtonGlyphs(self,PixelsPerInchOnDesign,Monitor.PixelsPerInch);
  end;

procedure TTimeSlotDialog.FormShow(Sender: TObject);
begin
  ShowTimeFields;
  end;

procedure TTimeSlotDialog.ShowTimeFields;
begin
  lbDays.Hide; reDays.Hide; udDays.Hide;
  FromDate.Format:=DpShortDateFmt;
  if rbTmNewer.Checked then begin
    lbFrom.Hide; lbTo.Hide;
    with FromDate do begin
      Top:=rbTmNewer.Top-4; Visible:=true;
      Kind:=dtkDate;
      end;
    ToDate.Visible:=false;
    end
  else if rbTmOlder.Checked then begin
    lbFrom.Hide; lbTo.Hide;
    FromDate.Visible:=false;
    with ToDate do begin
      Top:=rbTmOlder.Top-4; Visible:=true;
      end;
    end
  else if rbTmSpan.Checked then begin
    lbFrom.Visible:=true; lbTo.Visible:=true;
    with FromDate do begin
      Top:=lbFrom.Top-4; Visible:=true;
      Kind:=dtkDate;
      end;
    with ToDate do begin
      Top:=lbTo.Top-4; Visible:=true;
      end;
    end
  else if rbTmToday.Checked then begin
    lbFrom.Hide; lbTo.Hide;
    with FromDate do begin
      Top:=rbTmToday.Top-4; Visible:=true;
      Kind:=dtkTime;
      Format:=TpShortTimeFmt;
      end;
    ToDate.Visible:=false;
    end
  else begin
    lbFrom.Visible:=false; lbTo.Visible:=false;
    FromDate.Visible:=false; ToDate.Visible:=false;
    if not rbTmAll.Checked then begin
      lbDays.Visible:=true;
      reDays.Visible:=true;
      udDays.Visible:=true;
      if rbTmLastDays.Checked then begin
        reDays.Top:=rbTmLastDays.Top-4;
        udDays.Top:=rbTmLastDays.Top-4;
        lbDays.Top:=rbTmLastDays.Top;
        end
      else begin
        reDays.Top:=rbTmPrevDays.Top-4;
        udDays.Top:=rbTmPrevDays.Top-4;
        lbDays.Top:=rbTmPrevDays.Top;
        end;
      end;
    end;
  end;

procedure TTimeSlotDialog.rbTmClick(Sender: TObject);
begin
  ShowTimeFields;
  end;

procedure TTimeSlotDialog.FromDateChange(Sender: TObject);
begin
  if (FromDate.Kind=dtkDate) and (FromDate.Date>ToDate.Date) then ToDate.Date:=FromDate.Date;
  end;

procedure TTimeSlotDialog.ToDateChange(Sender: TObject);
begin
  if FromDate.Date>ToDate.Date then FromDate.Date:=ToDate.Date;
  end;

{------------------------------------------------------------------- }
function TTimeSlotDialog.Execute (APos : TPoint; var TimeSlot : TTimePeriod) : boolean;
begin
  AdjustFormPosition(Screen,self,APos);
  with TimeSlot do begin
    FromDate.Date:=Dat1;
    FromDate.Time:=DayTime;
    ToDate.Date:=Dat2;
    reDays.Value:=Days;
    case Mode of
    tmNewer    : rbTmNewer.Checked:=true;
    tmOlder    : rbTmOlder.Checked:=true;
    tmSpan     : rbTmSpan.Checked:=true;
    tmToday    : rbTmToday.Checked:=true;
    tmLastDays : rbTmLastDays.Checked:=true;
    tmPrevDays : rbtmPrevDays.Checked:=true;
    else rbTmAll.Checked:=true;
      end;
    rgTimeSelect.ItemIndex:=integer(TmRef);
    Result:=ShowModal=mrOK;
    if Result then begin
      Dat1:=DateOf(FromDate.Date);
      Dat2:=DateOf(ToDate.Date);
      DayTime:=MinuteOfTheDay(FromDate.Date)/MinsPerDay;
      Days:=reDays.Value;
      if rbTmNewer.Checked then Mode:=tmNewer
      else if rbTmOlder.Checked then Mode:=tmOlder
      else if rbTmSpan.Checked then Mode:=tmSpan
      else if rbTmToday.Checked then Mode:=tmToday
      else if rbTmLastDays.Checked then Mode:=tmLastDays
      else if rbtmPrevDays.Checked then Mode:=tmPrevDays
      else Mode:=tmAll;
      TmRef:=TTimeRef(rgTimeSelect.ItemIndex);
      end
    else Mode:=tmOff;
    end;
  end;

{------------------------------------------------------------------- }
function EditTimeSlot (APos : TPoint; var TimeSlot : TTimePeriod) : boolean;
begin
  if not assigned(TimeSlotDialog) then TimeSlotDialog:=TTimeSlotDialog.Create(Application);
  Result:=TimeSlotDialog.Execute(APos,TimeSlot);
  FreeAndNil(TimeSlotDialog);
  end;

{------------------------------------------------------------------- }
function GetTimeSlotAsString (const TimeSlot : TTimePeriod) : string;
begin
  with TimeSlot do begin
    if (Mode=tmAll) then Result:=rsAll
    else begin
      case Mode of
      tmAll       : Result:='';
      tmNewer     : Result:=TryFormat(rsAfter,[DateToStr(Dat1)]);
      tmOlder     : Result:=TryFormat(rsBefore,[DateToStr(Dat2)]);
      tmSpan      : Result:=TryFormat(rsSpan,[DateToStr(Dat1),DateToStr(Dat2)]);
      tmToday     : Result:=TryFormat(rsTodayFrom,[TimeToStr(DayTime)]);
      tmPrevDays  : Result:=TryFormat(rsOlderThan,[GetPluralString(rsDay,rsDays,Days)]);
      tmLastDays  : Result:=TryFormat(rsNewerThan,[GetPluralString(rsDay,rsDays,Days)]);
        end;
      if TmRef=trCreate then Result:=Result+'[C]'
      else if TmRef=trBoth then Result:=Result+'[B]';
      end;
    end;
  end;

end.
